@@ -32,9 +32,10 @@
 #include <gdk-pixbuf/gdk-pixbuf.h>
 
 typedef struct _MetaGroup MetaGroup;
+typedef struct _MetaStruts MetaStruts;
 
-typedef void (*MetaWindowForeachFunc) (MetaWindow *window,
-                                       void       *data);
+typedef gboolean (*MetaWindowForeachFunc) (MetaWindow *window,
+                                           void       *data);
 
 typedef enum
 {
@@ -49,6 +50,15 @@ typedef enum
   META_WINDOW_SPLASHSCREEN
 } MetaWindowType;
 
+struct _MetaStruts
+{
+  /* struts */
+  MetaRectangle left;
+  MetaRectangle right;
+  MetaRectangle top;
+  MetaRectangle bottom;
+};
+
 struct _MetaWindow
 {
   MetaDisplay *display;
@@ -94,6 +104,7 @@ struct _MetaWindow
   
   /* Whether we're maximized */
   guint maximized : 1;
+  guint maximize_after_placement : 1;
 
   /* Whether we're shaded */
   guint shaded : 1;
@@ -191,6 +202,9 @@ struct _MetaWindow
 
   /* Are we in the move_resize queue? */
   guint move_resize_queued : 1;
+
+  /* Are we in the update_icon queue? */
+  guint update_icon_queued : 1;
   
   /* Used by keybindings.c */
   guint keys_grabbed : 1;     /* normal keybindings grabbed */
@@ -207,10 +221,8 @@ struct _MetaWindow
    */
   guint calc_placement : 1;
 
-  /* Has nonzero struts */
-  guint has_struts : 1; 
-  /* Struts are from the _WIN_HINTS do not cover deal */
-  guint do_not_cover : 1;
+  /* Note: can be NULL */
+  MetaStruts *struts;
 
   /* Transient parent is a root window */
   guint transient_parent_is_root_window : 1;
@@ -221,7 +233,13 @@ struct _MetaWindow
 
   /* has a shape mask */
   guint has_shape : 1;
+
+  /* icon props have changed */
+  guint need_reread_icon : 1;
   
+  /* if TRUE, window was maximized at start of current grab op */
+  guint shaken_loose : 1;
+
 #ifdef HAVE_XSYNC
   /* XSync update counter */
   XSyncCounter update_counter;
@@ -263,12 +281,6 @@ struct _MetaWindow
   /* x/y/w/h here get filled with ConfigureRequest values */
   XSizeHints size_hints;
 
-  /* struts */
-  int left_strut;
-  int right_strut;
-  int top_strut;
-  int bottom_strut;
-  
   /* Managed by stack.c */
   MetaStackLayer layer;
   int stack_position; /* see comment in stack.h */
@@ -278,10 +290,14 @@ struct _MetaWindow
   int dialog_pipe;
 
   /* maintained by group.c */
-  MetaGroup *cached_group;
+  MetaGroup *group;
 };
 
-#define META_WINDOW_ALLOWS_MOVE(w)     ((w)->has_move_func && !(w)->maximized && !(w)->fullscreen)
+/* These differ from window->has_foo_func in that they consider
+ * the dynamic window state such as "maximized", not just the
+ * window's type
+ */
+#define META_WINDOW_ALLOWS_MOVE(w)     ((w)->has_move_func && !(w)->fullscreen)
 #define META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS(w)   ((w)->has_resize_func && !(w)->maximized && !(w)->fullscreen && !(w)->shaded)
 #define META_WINDOW_ALLOWS_RESIZE(w)   (META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS (w) &&                \
                                         (((w)->size_hints.min_width < (w)->size_hints.max_width) ||  \
@@ -298,7 +314,11 @@ void        meta_window_queue_calc_showing (MetaWindow  *window);
 void        meta_window_minimize           (MetaWindow  *window);
 void        meta_window_unminimize         (MetaWindow  *window);
 void        meta_window_maximize           (MetaWindow  *window);
+void        meta_window_maximize_internal  (MetaWindow    *window,
+                                            MetaRectangle *saved_rect);
 void        meta_window_unmaximize         (MetaWindow  *window);
+void        meta_window_make_above         (MetaWindow  *window);
+void        meta_window_unmake_above       (MetaWindow  *window);
 void        meta_window_shade              (MetaWindow  *window);
 void        meta_window_unshade            (MetaWindow  *window);
 void        meta_window_change_workspace   (MetaWindow  *window,
@@ -341,6 +361,8 @@ void        meta_window_fill_vertical       (MetaWindow  *window);
  */
 void        meta_window_queue_move_resize  (MetaWindow  *window);
 
+void        meta_window_update_struts      (MetaWindow  *window);
+
 /* this gets root coords */
 void        meta_window_get_position       (MetaWindow  *window,
                                             int         *x,
@@ -410,13 +432,19 @@ void meta_window_set_gravity (MetaWindow *window,
 void meta_window_handle_mouse_grab_op_event (MetaWindow *window,
                                              XEvent     *event);
 
+GList* meta_window_get_workspaces (MetaWindow *window);
+
 gboolean meta_window_visible_on_workspace (MetaWindow    *window,
                                            MetaWorkspace *workspace);
 
-/* Get minimum work area for all workspaces we're on */
-void meta_window_get_work_area (MetaWindow    *window,
-                                gboolean       for_current_xinerama,
-                                MetaRectangle *area);
+void meta_window_get_work_area_current_xinerama (MetaWindow    *window,
+                                                 MetaRectangle *area);
+void meta_window_get_work_area_for_xinerama     (MetaWindow    *window,
+                                                 int            which_xinerama,
+                                                 MetaRectangle *area);
+void meta_window_get_work_area_all_xineramas    (MetaWindow    *window,
+                                                 MetaRectangle *area);
+
 
 gboolean meta_window_same_application (MetaWindow *window,
                                        MetaWindow *other_window);
@@ -432,11 +460,14 @@ void meta_window_refresh_resize_popup (MetaWindow *window);
 
 void meta_window_free_delete_dialog (MetaWindow *window);
 
-void meta_window_foreach_transient (MetaWindow            *window,
-                                    MetaWindowForeachFunc  func,
-                                    void                  *data);
-gboolean meta_window_is_ancestor_of_transient (MetaWindow *window,
-                                               MetaWindow *transient);
+void     meta_window_foreach_transient        (MetaWindow            *window,
+                                               MetaWindowForeachFunc  func,
+                                               void                  *data);
+gboolean meta_window_is_ancestor_of_transient (MetaWindow            *window,
+                                               MetaWindow            *transient);
+void     meta_window_foreach_ancestor         (MetaWindow            *window,
+                                               MetaWindowForeachFunc  func,
+                                               void                  *data);
 
 gboolean meta_window_warp_pointer (MetaWindow *window,
                                    MetaGrabOp  grab_op);
@@ -455,4 +486,12 @@ gboolean meta_window_get_icon_geometry (MetaWindow    *window,
 
 const char* meta_window_get_startup_id (MetaWindow *window);
 
+void meta_window_recalc_features (MetaWindow *window);
+
+void meta_window_queue_update_icon (MetaWindow *window);
+
 #endif
+
+
+
+
